#include "AGILe/AGILe.hh"
#include "AGILe/Generator.hh"
#include <dlfcn.h>

namespace AGILe {


  // Declare getLibPath function (uses binreloc)
  const string getLibPath();


  namespace Loader {


    Log& getLog() {
      return AGILe::Log::getLog("AGILe.Loader");
    }

    
    // Struct for holding library names and versions
    struct GenLibInfo {
      GenLibInfo(string libname, string pkgname="", string pkgversion="") {
        this->libname = libname;
        this->pkgname = pkgname;
        this->pkgversion = pkgversion;
      }
      string libname, pkgname, pkgversion;
    };

    // Static library handles
    vector<void*> handles;
    void* agilehandle = 0;

    // Libraries required for each known generator.
    typedef map<string, vector<GenLibInfo> > GenLibs;

    // Generator version is kept globally for use by the 
    // Generator::setVersion(...) function.
    string version;


    // Initialise the dynamic loader
    void initialize() {
      // Nothing to do with raw dlopen
    }


    void* loadGenLib(const GenLibInfo& info, int dlmode=(RTLD_LAZY|RTLD_GLOBAL)) {
      // Make the proper system library name
      string libname = info.libname;
      // Put "lib" on the front if needed
      //if (name.find("lib") != 0) name = "lib" + libname;
      // Put lib suffix on the end if needed
      if (libname.find(SYSDSO) == string::npos) libname += SYSDSO;

      // Build up a list of search directories and platforms
      vector<string> dirs, platforms;
      char* env = 0;
      if (info.pkgname.empty()) {
        // If we're trying to load an AGILe interface...

        // Use the AGILe library install path
        dirs.push_back(getLibPath());

        // Use the current dir, too.
        dirs.push_back(".");      

        // And then the user's (non-system) library path
        env = getenv("LD_LIBRARY_PATH");
        if (env) dirs += split(env);

      } else {
        // If were loading a real generator library...

        // Use the AGILe gen path variable first
        env = getenv("AGILE_GEN_PATH");
        if (env) dirs += split(env);

        // Try the Genser area on CERN AFS 
        dirs.push_back("/afs/cern.ch/sw/lcg/external/MCGenerators");

        // Make a list of platform tags to try
        // This non-standard env variable can be specified by the user.
        env = getenv("LCG_PLATFORM");
        if (env) platforms.push_back(env);
        // CMTCONFIG specifies the LCG platform tag, which is hard to obtain
        // programmatically).
        env = getenv("CMTCONFIG");
        if (env) platforms.push_back(env);
        // Also try a fallback of the LCG tag, computed at build time.
        platforms.push_back(LCGTAG);
      }
      // Also try with no platform tag
      platforms.push_back("");

      // Try to load lib here
      for (vector<string>::const_iterator d = dirs.begin(); d != dirs.end(); ++d) {
        for (vector<string>::const_iterator p = platforms.begin(); p != platforms.end(); ++p) {
          string libpath = *d + "/" + info.pkgname + "/" + info.pkgversion + "/";
          if (!p->empty()) libpath += *p;
          if (!info.pkgname.empty()) libpath += "/lib";
          libpath += "/" + libname;
          // Remove duplicate slashes from path
          while (libpath.find("//") != string::npos) {
            libpath.replace(libpath.find("//"), 2, "/");
          }
          void* lib = dlopen(libpath.c_str(), dlmode);
          getLog() << Log::TRACE << "Trying to load: " << libpath << " -> " 
                                      << std::boolalpha << bool(lib) << " (" << lib << ")" << endl;

          if (lib) {
            getLog() << Log::TRACE << "Successful load from " << libpath << " ***" << endl;
            return lib;
          }
        }
      }
      return 0;
    }


    GenLibs _getAllGenLibInfo() {
      set<string> versions;
      GenLibs gens;
      GenLibInfo lhapdfGenInfo("libLHAPDF");

      // Fortran Herwig + variants
      versions.insert("6507");
      versions.insert("6510");
      for (set<string>::const_iterator v = versions.begin(); v != versions.end(); ++v) {
        vector<GenLibInfo> baselibs, reqlibs;
        baselibs.push_back(GenLibInfo("libherwig", "herwig", *v));
        void* pdflib = loadGenLib(lhapdfGenInfo);
        if (pdflib) {
          dlclose(pdflib);
          baselibs.push_back(lhapdfGenInfo);
        } else {
          baselibs.push_back(GenLibInfo("libherwig_pdfdummy", "herwig", *v));
        }
        baselibs.push_back(GenLibInfo("libHepMCfio"));
        //
        reqlibs = baselibs;
        reqlibs.push_back(GenLibInfo("libherwig_dummy", "herwig", *v));
        reqlibs.push_back(GenLibInfo("libAGILeFHerwig"));
        gens["FHerwig:" + *v] = reqlibs;
        //
        reqlibs = baselibs;
        reqlibs.push_back(GenLibInfo("libherwig_dummy", "herwig", *v));
        reqlibs.push_back(GenLibInfo("libjimmy", "jimmy", "4.31"));
        reqlibs.push_back(GenLibInfo("libAGILeFHerwigJimmy"));
        gens["FHerwigJimmy:" + *v] = reqlibs;
        //
        reqlibs = baselibs;
        reqlibs.push_back(GenLibInfo("libalpher", "alpgen", "2.1.3"));
        reqlibs.push_back(GenLibInfo("libAGILeAlpGenFHerwig"));
        gens["AlpGenFHerwig:" + *v] = reqlibs;
        //
        reqlibs = baselibs;
        reqlibs.push_back(GenLibInfo("libcharybdisher", "charybdis", "1.003"));
        reqlibs.push_back(GenLibInfo("libAGILeCharybdisFHerwig"));
        gens["CharybdisFHerwig:" + *v] = reqlibs;
      }
      versions.clear();

      // Fortran Pythia + variants
      versions.insert("6206");
      versions.insert("6326");
      versions.insert("6403");
      versions.insert("6404");
      versions.insert("6406");
      versions.insert("6408");
      versions.insert("6411");
      versions.insert("6412");
      versions.insert("6413");
      versions.insert("6414");
      versions.insert("6415");
      versions.insert("6416");
      versions.insert("6417");
      for (set<string>::const_iterator v = versions.begin(); v != versions.end(); ++v) {
        vector<GenLibInfo> baselibs, reqlibs;
        baselibs.push_back(GenLibInfo("libpythia6", "pythia6", *v));
        void* pdflib = loadGenLib(lhapdfGenInfo);
        if (pdflib) {
          dlclose(pdflib);
          baselibs.push_back(lhapdfGenInfo);
        } else {
          baselibs.push_back(GenLibInfo("libpythia6_pdfdummy", "pythia6", *v));
        }
        baselibs.push_back(GenLibInfo("libHepMCfio"));
        //
        reqlibs = baselibs;
        reqlibs.push_back(GenLibInfo("libpythia6_dummy", "pythia6", *v));
        reqlibs.push_back(GenLibInfo("libAGILeFPythia"));
        gens["FPythia:" + *v] = reqlibs;
        //
        reqlibs = baselibs;
        reqlibs.push_back(GenLibInfo("libalppyt", "alpgen", "2.1.3"));
        reqlibs.push_back(GenLibInfo("libAGILeAlpGenFPythia"));
        gens["AlpGenFPythia:" + *v] = reqlibs;
        //
        reqlibs = baselibs;
        reqlibs.push_back(GenLibInfo("libcharybdispyt", "charybdis", "1.003"));
        reqlibs.push_back(GenLibInfo("libAGILeCharybdisFPythia"));
        gens["CharybdisFPythia:" + *v] = reqlibs;
      }

      #ifdef HAVE_SHERPA
      // Sherpa
      {
        vector<GenLibInfo> reqlibs;
        reqlibs.push_back(GenLibInfo("SHERPA-MC/libSherpaMain", "sherpa", "1.1.0"));
        reqlibs.push_back(lhapdfGenInfo);
        reqlibs.push_back(GenLibInfo("libAGILeSherpa"));
        gens["Sherpa"] = reqlibs;
      }
      #endif

      #ifdef HAVE_CCPYTHIA
      // Pythia 8
      {
        vector<GenLibInfo> reqlibs;
        reqlibs.push_back(GenLibInfo("libpythia8", "pythia8", "1.107"));
        reqlibs.push_back(GenLibInfo("libhepmcinterface", "pythia8", "1.107"));
        reqlibs.push_back(lhapdfGenInfo);
        reqlibs.push_back(GenLibInfo("libAGILeCCPythia"));
        gens["CCPythia"] = reqlibs;
      }
      #endif

      #ifdef HAVE_CCHERWIG
      // Herwig++
      {
        vector<GenLibInfo> reqlibs;
        reqlibs.push_back(GenLibInfo("Herwig++/Herwig", "herwig++", "2.2.0"));
        reqlibs.push_back(lhapdfGenInfo);
        reqlibs.push_back(GenLibInfo("libAGILeCCHerwig"));
        gens["CCHerwig"] = reqlibs;
      }
      #endif

      return gens;
    }


    vector<string> getAvailableGens() {
      GenLibs genlibs = _getAllGenLibInfo();

      // Find which generators are present by trying to dlopen() them.
      vector<string> availableGens;
      vector<void*> loadedGens;
      for (GenLibs::const_iterator g = genlibs.begin(); g != genlibs.end(); ++g) {
        getLog() << Log::TRACE << "Trying to load libs for " << g->first << endl;
        // All listed libraries must be available for each generator to be considered useable.
        bool ok = true;
        for (vector<GenLibInfo>::const_iterator l = g->second.begin(); l != g->second.end(); ++l) {
          void* lib = loadGenLib(*l);
          if (lib) {
            loadedGens.push_back(lib);
          } else {
            ok = false;
            break;
          }
        }
        // Close libs.
        for (vector<void*>::reverse_iterator l = loadedGens.rbegin(); l != loadedGens.rend(); ++l) {
          getLog() << Log::TRACE << "Closing library " << *l << endl;
          dlclose(*l);
        }
        getLog() << Log::TRACE << "Unloaded gen libs" << endl;
        loadedGens.clear();
        // If all libs loaded, add to "available" list.
        if (ok) { availableGens.push_back(g->first); }
      }
      return availableGens;
    }


    void loadGenLibs(const string& genname) {
      const string fioName = "libHepMCfio";
      string agileName = "";
      string pdflibName = "";

      // Get version string from colon separated version string if given.
      string name = genname;
      version = "";
      const size_t colonpos = genname.rfind(":");
      if (colonpos != string::npos) {
        const size_t versionlength = genname.length() - colonpos - 1;
        name = genname.substr(0, colonpos);
        version = genname.substr(colonpos+1, versionlength);
      }

      GenLibs genlibs = _getAllGenLibInfo();
      if (genlibs.find(genname) == genlibs.end()) {
        throw runtime_error("Trying to load an unknown generator combination: " + genname);
      } else {
        const vector<GenLibInfo> libs = genlibs[genname];
        for (vector<GenLibInfo>::const_iterator l = libs.begin(); l != libs.end(); ++l) {
          handles.push_back(loadGenLib(*l));
        }
      }

      // Check and reset dlopen errors
      for (vector<void*>::const_iterator h = handles.begin(); h != handles.end(); ++h) {
        if (*h == 0) {
          throw runtime_error(string("Failed to load libraries: ") + dlerror());
        }
      }
      dlerror();
    }


    template <typename Fn>
    inline Fn evil_cast(void* ptr) {
      return reinterpret_cast<Fn>(reinterpret_cast<size_t>(ptr));
    }


    Generator* createGen() {
      // Load the factory and create an instance of the class
      void* makeGen = dlsym(agilehandle, "create");
      if (char* err = dlerror()) {
        throw runtime_error((string("Can't load creation factory: ") + err).c_str());
      }
      
      // Make the generator object.
      Generator* gen = evil_cast<Creator>(makeGen)();

      // Set the generator version for e.g. HepML dumps of generator state.
      gen->setVersion(version);
      
      return gen;
    }


    void destroyGen(Generator* gen) {
      //Destroyer* destroy_gen = (Destroyer*) dlsym(agilemod, "destroy");
      //if (dlerror()) {
      //  throw runtime_error("Can't load destruction factory");
      //}
      //destroy_gen(gen);
      delete gen;
    }


    // Close the libraries
    void finalize() {
      for (vector<void*>::reverse_iterator h = handles.rbegin(); h != handles.rend(); ++h) {
        if (*h != 0) dlclose(*h);
      }
    }


  }
}
